package ca.slashdev.eval;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.json.JSONObject;
import org.json.JSONTokener;

/**
 * Simple expression evaluator.  Expressions can contain literal integer values,
 * identifiers, bracketed expressions, and basic mathematical operators:
 * 
 *      +,-,*,/
 * 
 * Expression evaluation is performed from left to right, ignoring standard
 * mathematical operator precedence.
 * 
 * @author josh.kropf
 */
public final class Expression {
   public static void main(String[] args) throws Exception {
      if (args.length == 0) {
         System.out.println("specify expression");
         System.exit(1);
      }
      
      String expr = args[0];
      Map<String, Object> env = new HashMap<String, Object>();
      
      if( args.length > 1 ) {
         JSONObject obj = new JSONObject(new JSONTokener(args[1]));
         Iterator<String> i = obj.keys();
         while( i.hasNext() ) {
            String key = i.next();
            env.put(key, obj.get(key));
         }
      }
      
      System.out.printf("Result %d\n", eval(expr, env));
   }
   
   public static int eval(String expr, Map<String, Object> env) throws Exception {
      return new Expression(expr, env).eval();
   }
   
   public static int eval(String expr) throws Exception {
      return eval(expr, new HashMap<String, Object>());
   }
   
   private int _cursor = -1;
   private char[] _expr;
   private Map<String, Object> _env;
   
   private Expression(String expr, Map<String, Object> env) {
      _expr = expr.toCharArray();
      _env = env;
   }
   
   private int eval() throws Exception {
      Operation op = new Operation();
      char ch;
      
      while( (ch=peekChar()) != 0 ) {
         if( Expression.isLiteral(ch) ) {
            op.putOperand(nextLiteral());
         } else if( Expression.isIdentifier(ch) ) {
            String var = nextIdentifier();
            if( !_env.containsKey(var) )
               throw new Exception("identifier '"+var+"' is undefined");
            
            op.putOperand(Integer.parseInt(_env.get(var).toString()));
         } else if( Expression.isOperator(ch) ) {
            op.setOperator(nextChar());
         } else if( ch == '(' ) {
            nextChar();
            op.putOperand(eval());
         } else if( ch == ')') {
            nextChar();
            return op.getLvalue();
         } else {
            nextChar();
         }
      }
      
      return op.getLvalue();
   }
   
   private int nextLiteral() {
      StringBuffer sb = new StringBuffer();
      
      do {
         sb.append(nextChar());
      } while( peekChar() != 0 && Expression.isLiteral(peekChar()) );
      
      return Integer.parseInt(sb.toString());
   }
   
   private String nextIdentifier() {
      StringBuffer sb = new StringBuffer();
      
      do {
         sb.append(nextChar());
      } while( peekChar() != 0 && Expression.isIdentifier(peekChar()) );
      
      return sb.toString();
   }
   
   private char nextChar() {
      return _cursor < _expr.length? _expr[++_cursor] : 0;
   }
   
   private char peekChar() {
      return _cursor != _expr.length -1? _expr[_cursor+1] : 0;
   }
   
   private static boolean isIdentifier(char ch) {
      return ch == '_' || (ch >= 97 && ch <= 122) || (ch >= 65 && ch <= 90);
   }
   
   private static boolean isLiteral(char ch) {
      return Character.isDigit(ch);
   }
   
   private static boolean isOperator(char ch) {
      return ch == '+' || ch == '-' || ch == '*' || ch == '/';
   }
}

final class Operation {
   private boolean _hasLvalue;
   private int _lvalue;
   private char _operator = 0;
   
   public void setOperator(char operator) {
      _operator = operator;
   }
   
   public int getLvalue() {
      return _lvalue;
   }
   
   public void putOperand(int value) throws Exception {
      if( _hasLvalue ) {
         if( _operator == '+' ) {
            _lvalue += value;
         } else if( _operator == '-' ) {
            _lvalue -= value;
         } else if( _operator == '*' ) {
            _lvalue *= value;
         } else if( _operator == '/' ) {
            _lvalue /= value;
         } else {
            throw new Exception("invalid operator in operation");
         }
      } else {
         _lvalue = value;
         _hasLvalue = true;
      }
      
      _operator = 0;
   }
}